import numpy as N


def rebin(thedata, factor,conserve=None):
  """
Rebins an array into an array with the same aspect ratio, but with
dimensions that are larger or smaller by a factor of 'factor'. Increase
or decrease of dimensions is determined by the sign of 'factor'.

If the conserve keyword is set, flux is conserved, and the total of
the output array will be equal (barring rounding errors) to the total 
of the input array.

Works for 2D arrays; would require more smarts to make it work for 
higher dimensions. 
  """
  assert abs(factor) != 1.0, "Rebinning at the same scale wastes cpu!"

  if factor < 0: # make the array smaller
    oldshape=thedata.shape
    factor=abs(factor)
    work=N.zeros(oldshape,thedata.dtype)
    for i in range(0,oldshape[0],factor):
      for j in range(0,oldshape[1],factor):
        if (conserve is None):
          work[i,j]=thedata[i:i+factor,j:j+factor].mean(dtype=N.float64)
        else:
          work[i,j]=thedata[i:i+factor,j:j+factor].sum(dtype=N.float64)
    return work[::factor,::factor]

  elif (factor > 0): # make the array bigger
    work= N.repeat( N.repeat(thedata, factor, axis=0), factor, axis=1)
    if (conserve is None):
      return work
    else:
      return work/(factor*factor)
      
#-------------------------------------------------------------------
def embed(thedata, border, initval=None):
  """
Embeds an array in the center of a larger array. The size of the border
is given in pixels. The value of the border region is determined by
initval, if specified; otherwise zero.

Works for 2D arrays; would require more smarts to make it work for 
higher dimensions.

  """
  import types

  dims = thedata.shape  #Rebind for clarity
  #Handle symmetric or asymmetric border, based on type of argument
  if type(border) is types.IntType:
      border = (border, border)

  #Compute new dimensions.
  newdim=[2*border[0]+dims[0], 2*border[1]+dims[1] ]

  #Create the large array
  if (initval is None):
    work=N.zeros(newdim,thedata.dtype)
  else:
    work=N.zeros(newdim) + initval

  #Put the data into the middle of it, and return it.
  work[border[0]:border[0]+dims[0],border[1]:border[1]+dims[1]] = thedata
  return work

#-------------------------------------------------------------------

def stripborder(thedata, border):
  return thedata[border:-border, border:-border]


