#! /usr/bin/env python

import sys,os,math,tempfile
import pyfits
from pyraf import iraf
from iraf import images
import numpy as N
from tfit import mynum


class Grid:
    """ A grid of zones for which the registration is calculated and applied"""

    def __init__(self,imname,zonesize):
        """
        @param imname: Name of reference (lo-res) image
        @type imname: filename

        @param zonesize: size of registration zone in lores pixels
        @type zonesize: integer
        """
        self.zonesize=zonesize
        h=pyfits.open(imname)[0].header
        self.nx,self.ny = h['naxis1']/zonesize, h['naxis2']/zonesize
        self.dx = N.zeros((self.ny,self.nx),N.float32)
        self.dy = N.zeros((self.ny,self.nx),N.float32)
        self.lx_set = range(0,h['naxis1'],zonesize)
        self.ly_set = range(0,h['naxis2'],zonesize)
        self._idx = [(x,y) for x in range(self.nx) for y in range(self.ny)]
        self.cells=[]

    def __iter__(self):
        return self._idx.__iter__()

    def addcell(self,cell):
        """ add the cell to the cell list, but also add its (dx,dy) to
        the dx,dy arrays """

        self.cells.append(cell)
        self.dx[cell.gridcell]=cell.dx
        self.dy[cell.gridcell]=cell.dy

    def pre_interp(self,border):
        """ Embed the dx/dy arrays in a border in preparation for
        using them to interpolate """
        self.border=border
        self._dx = mynum.embed(self.dx,border)
        self._dy = mynum.embed(self.dy,border)

    def maxshift(self):
        """ Return the maximum shift in all cells.."""
        mdx=self.dx.max()
        mdy=self.dy.max()
        mdr = math.sqrt(mdx*mdx + mdy*mdy)
        return mdx, mdy, mdr
    
class Zone:
    """ A single zone in which to compute and apply the misregistration
    correction """

    #Define some globals
    #Filenames (for input)
    imname=None
    refname=None
    Psfname=None
    #File handles (for output)
    reglog=None
    shiftlog=None
    lut=None
    diag=None
    #Shift tolerance and scale ratio parameters
    tolerance=None
    scale_ratio=None

    def __init__(self,xr,yr,grid):
        """ Instantiate a zone based on its position in a L{Grid}

        @param xr,yr: indices into the grid array
        @type xr,yr: integer

        @param grid
        @type grid: L{Grid}
        """

        self.gridcell=(yr,xr)
        self.xy=(xr,yr)

        self.lx = grid.lx_set[xr]
        self.ly = grid.ly_set[yr]
 
        self.ux = self.lx + grid.zonesize
        self.uy = self.ly + grid.zonesize
        
        self.cx = self.lx + (self.ux - self.lx)/2.0
        self.cy = self.ly + (self.uy - self.ly)/2.0
        self.range="[%d:%d,%d:%d]"%(self.lx,self.ux,self.ly,self.uy)

        self.dx, self.dy = (0,0)
        self.psfname = self.Psfname
        self.interpolated = False
        
    def xreg(self):
        """ Perform the IRAF xregister task on this zone, and fill in
        the self.dx, self.dy attributes with the result. """

        #Do the registration
        tmpfile=tempfile.mktemp('tmp','.dance')
        iraf.xregister(input=self.imname,
                       reference=self.refname,
                       regions=self.range,
                       databasefmt=iraf.yes,
                       shifts=tmpfile,
                       append=iraf.no,
                       lorej=3,
                       hirej=3,
                       background="none",
                       verbose=iraf.no,
                       Stderr=self.dancelog)
        xshift,yshift=parse_xreg_output(tmpfile)
        os.remove(tmpfile)

        #Save the result if it passes tolerance. Note scale_ratio comes into play here.
        if ((abs(xshift) <= self.tolerance) and (abs(yshift) <= self.tolerance)):
            self.dx,self.dy = xshift*self.scale_ratio,yshift*self.scale_ratio
            
        #Always write out the log with the raw shifts
        self.reglog.write("%s \t%f %f\n"%(self.range,xshift,yshift))


    def interp(self,grid):
        """ If this zone's (dx,dy) is zero, interpolate its value
        from its neighbors """

        _yr, _xr = self.gridcell[0] + grid.border, self.gridcell[1] + grid.border

        if self.dx == 0:
            try:
                sub=grid._dx[_yr-1:_yr+2, _xr-1:_xr+2]
                idx=N.where(sub != 0)
                ans=sub[idx].mean(dtype=N.float64)
                if ans <= self.tolerance:
                    self.dx = ans
                    grid.dx[self.gridcell] = ans
                    self.interpolated = True
            except ZeroDivisionError:
                pass #ok, we're on edge of range

        if self.dy == 0:
            try:
                sub=grid._dx[_yr-1:_yr+2, _xr-1:_xr+2]
                idx=N.where(sub != 0)
                ans=sub[idx].mean(dtype=N.float64)
                if ans <= self.tolerance:
                    self.dy = ans
                    grid.dy[self.gridcell] = ans
                    self.interpolated=True

            except ZeroDivisionError:
                pass #ok, we're on edge of range

    def shiftpsf(self,psf):
        """ If the shift is nonzero, apply it to the psf using the
        IRAF task imlintran.
        Update the lookup table in any case. """

        if (self.dx != 0 or self.dy != 0):
            #shift the psf
            self.psfname = psf.name.replace(".fits","_%02d_%02d.fits"%self.xy)
            iraf.imlintran(xin=psf.xcent,
                           yin=psf.ycent,
                           xout=psf.xcent + self.dx,
                           yout=psf.ycent + self.dy,
                           input=psf.name,
                           output=self.psfname,
                           xmag=1.0, ymag=1.0,
                           xrot=0.0,yrot=0.0,
                           ncols='INDEF', nlines='INDEF',
                           interp='drizzle',
                           bound='constant',
                           constant=0.0,
                           mode='h',
                           Stdout=self.shiftlog)

        #in any case, update the files
        self.lut.write("%d  %d  %d  %d  %s\n"%(self.lx,self.ux,self.ly,self.uy,self.psfname))
        self.diag.write("%d %d %f %f %s %d\n"%(self.cx,self.cy,self.dx,self.dy,self.range,self.interpolated))

class PSF:
    def __init__(self,psfname):
        self.name=psfname
        h=pyfits.open(self.name)[0].header
        self.nx,self.ny=h['naxis1'],h['naxis2']
        self.xcent=self.nx/2 + 0.5
        self.ycent=self.ny/2 + 0.5
                      
def parse_xreg_output(tmpfile):
    results=open(tmpfile).readlines()
    dx=float(results[-2].split()[-1])
    dy=float(results[-1].split()[-1])
    
    return dx,dy

def writeheader(fh,colnames):
    """Write an SExtractor-style header for these column names"""
    for i in range(len(colnames)):
        fh.write("#  %d    %s \n"%(i+1,colnames[i]))
        
def main(refname,imname,psfname,zonesize,tolerance,scale_ratio,lutname,diagname,logname):
    #First define the globals
    Zone.refname=refname
    Zone.imname=imname
    Zone.Psfname=psfname
    Zone.tolerance=tolerance
    Zone.scale_ratio=scale_ratio

    #includes opening some file handles
    Zone.lut=open(lutname,'w')
    Zone.diag=open(diagname,'w')
    Zone.shiftlog=open(logname+'_shift','w')
    Zone.reglog=open(logname+'_reg','w')
    Zone.dancelog=open(logname+'_dance','w')

    writeheader(Zone.diag,["cx","cy","xshift","yshift","range","interpolated"])

    #Define the grid
    grid=Grid(imname,zonesize)

    #Iterate over the grid to do the cross-correlation
    for (xr,yr) in grid:
        cell=Zone(xr,yr,grid)
        cell.xreg()
        grid.addcell(cell)

        
    count = 0
    #Now fill in the holes if necessary, and shift the psf
    grid.pre_interp(2)
    psf=PSF(psfname)
    for cell in grid.cells:
        cell.interp(grid)
        cell.shiftpsf(psf)

    #close all the files
    Zone.lut.close()
    Zone.diag.close()
    Zone.shiftlog.close()
    Zone.reglog.close()
    Zone.dancelog.close()

    #Return the maximum shift
    mdx,mdy,mdr=grid.maxshift()
    return "Max shift dx, dy, dr = %f, %f, %f lores pixels"%(mdx/scale_ratio,
                                                             mdy/scale_ratio,
                                                             mdr/scale_ratio)

def test():
    os.system('rm tfd*')
    main(' ../n_irac_1_s1_v0.31_sci_bgcorr2.fits',
         '../out_nch1_ep1_d1.9z_collage.fits',
         'ch1psf_10x_rothof_adj.fits',
         100, 1.0, 10,
         'tfd_lookup.txt',
         'tfd_diag.txt',
         'tfd_log.txt')

if __name__ == '__main__':
    import sys, futil
    if len(sys.argv) == 1:
        test()
    elif len(sys.argv) == 2:
        d=futil.get_pars(sys.argv[1])
        main(d['LORESIMAGE'],
             d['MODELIMAGE'],
             d['PSFIMAGE'],
             int(d['ZONESIZE']),
             float(d['MAXSHIFT']),
             int(d['REL_SCALE']),
             d['LOOKUPFILE'],
             d['DIAGFILE'],
             d['LOGFILE'])
    else:
        p1,p2,p3,p4,p5,p6,p7,p8,p9=sys.argv[1:]
        main(p1,p2,p3,p4,p5,p6,p7,p8,p9)
