import os, glob, time, copy, warnings
import pyfits
import numpy as N
from tfit import fitsimage, box, template, mynum
from template import MaxFlagError
from box import NonOverlapError
True  = 1
False = 0

def process(Indir, lores_file, PSF_file, scale_factor, outdir, fake=None,
            flagfile=None,maxflag=0,savehi=False,hr_outdir=None):
  """
    #Module to process image templates, stored in FITS files.
    #Limiting assumptions:
    #   Pixels must be aligned on the same grid (a la master GOODS grid)
    #   Pixels must be square
    #   The templates are of a higher resolution & need to be convolved down
    #   The PSF array is square, and at the same pixel resolution as the templates.
    #   The PSF array has been normalized to 1.
    #   The scale factor is an even integer, or 1.
 
  """
  assert os.path.isdir(Indir), "Process: Input directory is not a directory"
  assert os.path.exists(PSF_file), "Process: PSF file does not exist"
  assert type(scale_factor) == type(5), "Process: Scale factor must be an integer"

  #TTD: scale factor can be derived from headers.
  #     do more testing

##   print "This is the version with the flag check"
##   print "Flagfile %s"%flagfile
##   print "maxflag is %d"%maxflag
  
  if not os.path.isdir(outdir):
    os.makedirs(outdir)

  if savehi and not os.path.isdir(hr_outdir):
    os.makedirs(hr_outdir)

  # Get the PSF array
  f1=pyfits.open(PSF_file)
  psf = f1[0].data
  # Normalize it, if it hasn't been
  if psf.sum() != 1.0:
    psf = psf/psf.sum(dtype=N.float64)
#  border = max(psf.shape) #+0.5 #to push to the pixel corner
  border = (max(psf.shape) + 1)/2 #We don't really need the *full* size for this
  f1.close()

  #Get the lores file: needed for its header
  lores = fitsimage.Image(lores_file)
  lores.wcs.fillcd()

  #Get the lores flag image, if required:
  if flagfile:
    try:
      flag=fitsimage.Image(flagfile)
    except:
      flag=None
  else:
    flag=None

  #Suppress annoying warnings from pyfits:
  warnings.simplefilter('ignore',UserWarning)
  # Get the list of templates
  tlist=glob.glob(Indir+'/*.fits')
  # Begin processing the list
  now=time.time()
  cpunow=time.clock()
  pcount=0
  for obj in tlist:
    newname = outdir+'/'+ os.path.basename(obj)
    if os.path.isfile(newname):
      print "Skipping %s : template already exists" % obj
    else:
      print time.asctime(),": Processing "+obj+"..."
      
      #Get the image information.
      hicut = fitsimage.Image(obj)
      #Spiffen up the wcs
      hicut.wcs.fillcd()
      #Do the real work for this image:
      try:
        locut = template.blur_one(hicut, lores, border, psf, scale_factor,
                                  fake,flag,maxflag)

        #Now mess with the filename bits
        locut.filename = newname

        #Add the name of the psf file to its header
        locut.h.update('CONVKERN',PSF_file,comment='Convolution kernel used to produce template')
        locut.h.update('TBINFACT',scale_factor,comment='Template binning')

        #And write the file.
        locut.writeto(locut.filename)
        pcount+=1
        
        #Save the hires if desired
        if savehi and scale_factor != 1:
          #Undo the binning
          newimg=mynum.rebin(locut.d,scale_factor,conserve='true')
          #Make a new FITSimage with the correct values
          htwcs=copy.deepcopy(locut.wcs)
          htwcs.rescale(-1*scale_factor)
          hrt=template.assemble_template(locut,newimg,htwcs)
          hrt.h.update('TBINFACT',1,comment='Template binning')
          #write the file
          hrt.writeto(locut.filename.replace(outdir,hr_outdir))

                        

      except MaxFlagError:
        print "%s flagged out: omitting"%obj
      except NonOverlapError:
        print "%s falls outside lores image: omitting"%obj

  #turn warnings back on
  warnings.simplefilter('always',UserWarning)
  #Return the final count
  return "%d objects processed"%pcount
#---------------------------------------------------------------------------

def cull(LoResFile,HiResFile,catfile,indir):
    assert os.path.isdir(indir), "Cull: Input directory %s is not a directory"%indir
    assert os.path.exists(LoResFile), "Cull: LoResFile %s does not exist"%LoResFile
    assert os.path.exists(HiResFile), "Cull: HiResFile %s does not exist"%HiResFile
    assert os.path.exists(catfile), "Cull: catfile %s does not exist"%catfile

    #TTD: insert use of lores weightmap, to set weight_has_good_data

    #Get the lores header
    LoRes = fitsimage.Image(LoResFile)
    LoRes.wcs.fillcd()

    #Get the hires header
    #... modification to permit multi-tile catalogs
    if HiResFile.lower().endswith('fits'):
      hwcs = fitsimage.Image(HiResFile).wcs
      hwcs.fillcd()
    else: #it's a text file. Read its info.
      hf=open(HiResFile).readlines()
      wdict={}
      for line in hf:
        key,value=line.split()
        wdict[key.upper()]=float(value)
      #Start with the lores wcs
      hwcs=copy.deepcopy(LoRes.wcs)

      #set the scaling
      hwcs.rescale(-1*wdict['REL_SCALE'])

      #Replace the crpixes
      hwcs.set_crpix(wdict['CRPIX1'],wdict['CRPIX2'])

                     
    weight_has_good_data = 1  #Insert WtMapFile,WtMap later

    #Read the catalog
    oldcat=open(catfile,'r').readlines()

    #Start some bookkeeping lists
    newcat = []
    rmlist = []
    plist = []

    

    #Go through each object & determine whether to keep as is, change, or cull
    #Do the bookkeeping and take the appropriate actions for each object.
    for line in oldcat:
	if not line.startswith('#'):
	    eliminate=False
	    change = False
	    id,x,y,rest = line.split(None,3)
	    t=indir+'/'+id+'.fits'
            
            try:
              thisone = fitsimage.Image(t)
              if LoRes.wcs_box.Contains(thisone.wcs_box):
                  pass
              else:
                  if (LoRes.wcs_box.Overlaps(thisone.wcs_box) or
                      thisone.wcs_box.Overlaps(LoRes.wcs_box)): #can happen!
                  
                      print "Overlap detected: id ",id
                     # print thisone.wcs_box
                      thisone.ShrinksToFit(LoRes)
                      plist.append(os.path.basename(t))
                      change=True
                  else:
                      eliminate=True

      ## 	sub = WtMap.xypos(thisone)
      ## 	weight = WtMap.d[sub.Slice()]

              if (weight_has_good_data):
                  if change:
                      newname=os.path.dirname(t)+'/partial_'+os.path.basename(t)
                      thisone.writeto(newname)
              else:
                  eliminate=True

              if not eliminate:
                  nx,ny = LoRes.wcs.xy2xy(float(x),float(y),hwcs)
                  #Push the object center into the lores range
                  #Flag the objects for which this moves them.
                  nx, xf = template.clip(LoRes.local_box.Xrange(),nx) 
                  ny, yf = template.clip(LoRes.local_box.Yrange(),ny) 
                  nline=' '.join([id,`float(nx)`,`float(ny)`,rest.rstrip(),`xf+yf`,"\n"])
                  newcat.append(nline)

              else:
                  rmlist.append(t)
                  
            except IOError: #Permits using a catalog containing objects not cutout
              rmlist.append(t)
    
    #Tell us the final counts
    print "Size of catalog to keep  : ",len(newcat)
    print "Number of partial files  : ",len(plist)
    print "Number of files to remove: ",len(rmlist)

    #Now write out the new catalog
    #os.rename(catfile,catfile+'_old')
    f=open(catfile+'_culled','w')
    for line in newcat:
	f.write(line)
    f.close()

    #Also clean up the "partial" files
    if len(plist) > 0:
      try:
	os.mkdir(indir+'/shrunk')
      except OSError,e:
        if e.strerror == 'File exists':
          pass
        else:
          raise e
        
      for thing in plist:
          oldname=indir+'/'+thing
          os.rename(oldname,indir+'/shrunk/full_'+thing)
          os.rename(indir+'/partial_'+thing,indir+'/'+thing)

    #and "remove" the fits files
    if len(rmlist) > 0:
      try:
	os.mkdir(indir+'/culled')
      except OSError,e:
        if e.strerror == 'File exists':
          pass
        else:
          raise e

      for thing in rmlist:
        if os.path.isfile(thing):
          newname=os.path.dirname(thing)+'/culled/'+os.path.basename(thing)
          os.rename(thing,newname)


def uncull(templatedir):
  """ Restore the original template directory """

  plist=glob.glob(templatedir+'/shrunk/*.fits')
  if len(plist) > 0:
    for thing in plist:
      newname=thing.replace('/shrunk/full_','')
      shutil.copyfile(thing,newname)
    os.rmdir(templatedir+'/shrunk/')

  dlist=glob.glob(templatedir+'/culled/*.fits')
  if len(dlist) > 0:
    for thing in dlist:
      shutil.copyfile(thing,templatedir)
    os.rmdir(templatedir+'/culled/')

def multikern(gridfile,indir,outdir,tfit_cat,loresfile,scale_factor,posframe='lowres',flagfile=None,maxflag=0):
  """ Process cutouts through different kernels based on their positions"""

  #Read the gridfile into a dictionary keyed by boxes, values = psf file
  psfgrid=parsegrid(gridfile)

  #Read in the lores file for its wcs
  LoRes=fitsimage.Image(loresfile)

  #Get the lores flag image, if required:
  if flagfile:
    flag=fitsimage.Image(flagfile)
  else:
    flag=None

  #Open the catalog file
  f=open(tfit_cat)

  #Parse which reference frame the catalog positions are in
  posframe=posframe.lower()
  if not posframe.startswith('hi') and not posframe.startswith('lo'):
    raise ValueError,"Illegal value for posframe: legal values are [hires, lores]"
  #Create  the output directory if necessary.
  if not os.path.isdir(outdir):
    os.mkdir(outdir)

  #Suppress annoying warnings from pyfits:
  warnings.simplefilter('ignore',UserWarning)
    
  #Bookkeeping
  notfound={}
  count=0
  catcount=0
  #Now walk through the catalog.
  for line in f:
    if not line.startswith('#'):
      id,x,y,rest = line.split(None,3)
      t=indir+'/'+id+'.fits'
      catcount+=1

      #Transform to lores coords if necessary
      if posframe.startswith('hi'):
        nx,ny = LoRes.wcs.xy2xy(float(x),float(y),hicut.wcs)
      else:
        nx,ny = int(round(float(x))),int(round(float(y)))

      #Find the right psf
      found=None
      for k in psfgrid:
        if k.Includes(nx,ny):
          found=True
          newname = outdir+'/'+ os.path.basename(t)
          if template_exists(newname,psfgrid[k]):
            print "Template with correct PSF already exists for %s: skipping"%id
          else:
            print "Starting work on %s"%id
            #Found it. Start work.
            thispsf,border=readpsf(psfgrid[k])
            hicut=fitsimage.Image(t)
            try:
              locut=template.blur_one(hicut,LoRes,border,thispsf,scale_factor,
                                      None,flag,maxflag)
              #Now mess with the filename bits
              locut.filename=newname
              #Add the name of the psf file to its header
              locut.h.update('CONVKERN',psfgrid[k],
                             comment='Convolution kernel used to produce template')
              #And write the file.... wow, this works...
              locut.writeto(locut.filename)
              count+=1
              #break out of the loop
              break
            except MaxFlagError:
              print "%s flagged out: omitting"%id
            except NonOverlapError:
              print "%s falls entirely outside lores image: omitting"%id

      if not found:
        print "No PSF file found for %s, x/y = %f,%f"%(t,nx,ny)
        notfound[t]=(nx,ny)
  #That's it.
  f.close()

  #Report errors
  if len(notfound) > 0:
    print "%d templates could not be found in the lookup table"%len(notfound)
    out=open(gridfile+'_notfound','w')            
    for k in notfound:
      print >> out,k,":",notfound[k]
    out.close()
  else:
    print "\n==== Multikern Results ===="
    print "All templates were successfully located in the lookup table."
    print "%d templates of %d were successfully convolved\n"%(count,catcount)

  #Clean up any old templates in case they exist
  if count > 0:
    try:
      os.mkdir(outdir+'/old/')
      mvlist=glob.glob('%s/*_old.fits'%outdir)
      for name in mvlist:
        os.rename(name,"%s/old/%s"%(outdir,os.path.basename(name)))

    except OSError:
      "Error cleaning up old files: should be done by hand"

  #turn warnings back on
  warnings.simplefilter('always',UserWarning)

  return "Of %d templates, %d not found, %d omitted, %d convolved"%(catcount, len(notfound), catcount-count, count)

            
def readpsf(PSF_file):
    # Get the PSF array
  f1=pyfits.open(PSF_file)
  psf = f1[0].data
  # Normalize it, if it hasn't been
  if psf.sum() != 1.0:
    psf = psf/psf.sum(dtype=N.float64)
#  border = max(psf.shape) #+0.5 #to push to the pixel corner
  border = (max(psf.shape) + 1)/2 #We don't really need the *full* size for this
  f1.close()
  return psf,border

def findbin(lookup,x,y):
  """ Find the bin containing an object. """
  for k in lookup:
    if k.Includes(x,y):
      return(lookup[k])

#Gridfile should contain x1 x2 y1 y2 psffilename
#so to parse this:
def parsegrid(gridfile):
  d={}
  f=open(gridfile)
  for line in f:
    x1,x2,y1,y2,kern=line.split()
    k=box.Box(x1=int(float(x1)),y1=int(float(y1)),x2=int(float(x2)),y2=int(float(y2)))
    d[k]=kern
  f.close()
  return d

             
def template_exists(tname,psfname):
  """ Check not only whether a file of the correct name exists,
  but also whether it was created with a PSF of the correct name """
  try:
    f=pyfits.open(tname)
    convkern=f[0].header['convkern']
    if os.path.basename(psfname).lower() == convkern.lower():
      found = True
      #print "template exists with correct kernel"
    else:
      found = False
      os.rename(tname,tname.replace('.fits','_old.fits'))
      #print "template exists with wrong kernel; renamed"
  except IOError:
    found = False

  return found
